use serde::{Serialize, Deserialize};
use crate::{
    models::item::{SkillType, Item}
};
use once_cell::sync::OnceCell;
use std::collections::BTreeMap;
use rand::seq::SliceRandom;
use super::species::BodyPart;

pub type AttackMessageChoice = Vec<Box<dyn Fn(&Item, &Item, bool) -> String + 'static + Sync + Send>>;
pub type AttackMessageChoicePart = Vec<Box<dyn Fn(&Item, &Item, &BodyPart, bool) -> String + 'static + Sync + Send>>;

pub struct SkillScaling {
    pub skill: SkillType,
    pub min_skill: f64,
    pub mean_damage_per_point_over_min: f64
}

pub struct WeaponData {
    pub uses_skill: SkillType,
    pub raw_min_to_learn: f64,
    pub raw_max_to_learn: f64,
    pub normal_attack_start_messages: AttackMessageChoice,
    pub normal_attack_success_messages: AttackMessageChoicePart,
    pub normal_attack_mean_damage: f64,
    pub normal_attack_stdev_damage: f64,
    pub normal_attack_skill_scaling: Vec<SkillScaling>,
}

impl Default for WeaponData {
    fn default() -> Self {
        Self {
            uses_skill: SkillType::Blades,
            raw_min_to_learn: 0.0,
            raw_max_to_learn: 15.0,
            normal_attack_start_messages:
            vec!(Box::new(|attacker, victim, exp| format!(
                "{} makes an attack on {}",
                &attacker.display_for_sentence(exp, 1, true),
                &victim.display_for_sentence(exp,1, false)))),
            normal_attack_success_messages:
            vec!(Box::new(|attacker, victim, part, exp|
                          format!("{}'s attack on {} hits {} {}",
                                  &attacker.display_for_sentence(exp, 1, true),
                                  &victim.display_for_sentence(exp, 1, false),
                                  &victim.pronouns.possessive,
                                  part.display(victim.sex.clone())
                          ))),
            normal_attack_mean_damage: 1.0,
            normal_attack_stdev_damage: 2.0,
            normal_attack_skill_scaling: vec!()
        }
    }
}

pub struct PossessionData {
    pub weapon_data: Option<WeaponData>
}

impl Default for PossessionData {
    fn default() -> Self {
        Self {
            weapon_data: None
        }
    }
}

impl WeaponData {
    pub fn normal_attack_start_message(
        &self,
        attacker: &Item, victim: &Item, explicit_ok: bool) -> String {
        let mut rng = rand::thread_rng();
        self.normal_attack_start_messages[..].choose(&mut rng).map(
            |f| f(attacker, victim, explicit_ok)).unwrap_or(
            "No message defined yet".to_owned())
    }

    pub fn normal_attack_success_message(
        &self, attacker: &Item, victim: &Item,
        part: &BodyPart, explicit_ok: bool
    ) -> String {
        let mut rng = rand::thread_rng();
        self.normal_attack_success_messages[..].choose(&mut rng).map(
            |f| f(attacker, victim, part, explicit_ok)).unwrap_or(
            "No message defined yet".to_owned())
    }

    
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum PossessionType {
    // Special values that substitute for possessions.
    Fangs, // Default weapon for certain animals
    // Real possessions from here on:
}

pub fn fist() -> &'static WeaponData {
    static FIST_WEAPON: OnceCell<WeaponData> = OnceCell::new();
    FIST_WEAPON.get_or_init(|| {
        WeaponData {
            uses_skill: SkillType::Fists,
            raw_min_to_learn: 0.0,
            raw_max_to_learn: 2.0,
            normal_attack_start_messages: vec!(
                Box::new(|attacker, victim, exp|
                         format!("{} swings at {} with {} fists",
                                 &attacker.display_for_sentence(exp, 1, true),
                                 &victim.display_for_sentence(exp, 1, false),
                                 &attacker.pronouns.possessive
                         )
                )
            ),
            normal_attack_success_messages: vec!(
                Box::new(|attacker, victim, part, exp|
                         format!("{}'s fists smash into {}'s {}",
                                 &attacker.display_for_sentence(exp, 1, true),
                                 &victim.display_for_sentence(exp, 1, false),
                                 &part.display(victim.sex.clone())
                         )
                )
            ),
            ..Default::default()
        }
    })
}

pub fn possession_data() -> &'static BTreeMap<PossessionType, PossessionData> {
    static POSSESSION_DATA: OnceCell<BTreeMap<PossessionType, PossessionData>> = OnceCell::new();
    use PossessionType::*;
    &POSSESSION_DATA.get_or_init(|| {
        vec!(
            (Fangs, PossessionData {
                weapon_data: Some(WeaponData {
                    uses_skill: SkillType::Fists,
                    raw_min_to_learn: 0.0,
                    raw_max_to_learn: 2.0,
                    normal_attack_start_messages: vec!(
                        Box::new(|attacker, victim, exp|
                                 format!("{} bares {} teeth and lunges at {}",
                                         &attacker.display_for_sentence(exp, 1, true),
                                         &attacker.pronouns.possessive,
                                         &victim.display_for_sentence(exp, 1, false),
                                 )
                        )
                    ),
                    normal_attack_success_messages: vec!(
                        Box::new(|attacker, victim, part, exp|
                                 format!("{}'s teeth connect and tear at the flesh of {}'s {}",
                                         &attacker.display_for_sentence(exp, 1, true),
                                         &victim.display_for_sentence(exp, 1, false),
                                         &part.display(victim.sex.clone())
                                 )
                        )
                    ),
                    ..Default::default()
                }),
                ..Default::default()
            })
        ).into_iter().collect()
    })
}
